(* Compilation from source expressions. *)

structure Compile :> Compile = struct

open Data;
open Eqset; 
infix 6 ++ --

(* Normalization, translation to de Bruijn indices, and generation
   of environment trimmers.  Assumes that a variable name cannot be
   the empty string.  *)

fun compile e = 
    let infixr seq 
        exception UnboundVariable
        fun lookup ([], x)    = raise UnboundVariable
          | lookup (y::yr, x) = if x=y then 1 else 1 + lookup(yr, x)
        fun range (i,j) = if i>j then [] else i :: range(i+1,j)
        fun maketrim (ce, vars) = 
            let fun h (i, [],    ce', ts) = (rev ce', rev ts)
                  | h (i, x::xr, ce', ts) = 
                    if mem(x,vars) then h(i+1, xr, x::ce', i::ts)
                                   else h(i+1, xr, ce',    ts)
            in h(1, ce, [], []) end
        fun trimexp ce (co, li) =
            let val (ce', ts) = maketrim (ce, li)
            in (co ce', ts) end
        fun livein colis = union (map (fn (co, li) => li) colis)
        fun prim e = (fn ce => e, emptyset)
        fun (co1, li1) seq (co2, li2) =
            (fn ce => let val (e2, t) = trimexp ce (co2, li2)
                      in Case (co1 ce, ([(e2, range(1, length t))], t)) end,
             li1 ++ li2)
        fun compilealts alts =
            let val yscolis = map (fn (k, ys, e) => (ys, comp e)) alts
                val live = union (map (fn (ys,(_,li)) => li--setof ys) yscolis)
                fun compalt ce (ys, (co, li)) = trimexp (rev ys @ ce) (co, li)
                fun buildalts ce =
                    let val (ce', t) = maketrim (ce, live)
                    in (map (compalt ce') yscolis, t) end
            in (buildalts, live) end
        and comp (VarS x)            = (fn ce => Var (lookup(ce, x)), set x)
          | comp (AppS(e1, VarS x))  = 
            let val (co1, li1) = comp e1
            in (fn ce => App (co1 ce, lookup(ce, x)), li1 ++ set x) end
          | comp (AppS(e1, e2))      = comp (LetS([("",e2)],AppS(e1,VarS "")))
          | comp (LetS(binds, e0))   =
            let val colis               = map (comp o #2) binds
                val coli0 as (co0, li0) = comp e0
                val names               = map #1 binds
                fun buildlet ce =
                    let val ce'  = names @ ce
                        val ce'' = map (fn _ => "") binds @ ce
                    in Let (map (trimexp ce'') colis, trimexp ce' coli0) end
                
            in (buildlet, livein colis ++ (li0 -- setof names)) end
          | comp (LetrecS(binds, e0)) =
            let val colis               = map (comp o #2) binds
                val coli0 as (co0, li0) = comp e0
                val names               = map #1 binds
                fun buildlet ce =
                    let val ce' = names @ ce
                    in Let (map (trimexp ce') colis, trimexp ce' coli0) end
            in (buildlet, (livein colis ++ li0) -- setof names) end
          | comp (LamS(x, e))        = 
            let val (co, li) = comp e
            in (fn ce => Lam (co (x :: ce)), li -- set x) end
          | comp (ConS(k, []))       = (fn ce => Con(k, 0), emptyset)
          | comp (ConS(k, es))       = 
            let val colis = map comp (rev es)
                val a     = length es
                fun buildlet ce =
                    let val ce'' = map (fn _ => "") es @ ce
                    in Let (map (trimexp ce'') colis, 
                            (Con(k,a), range(1,a))) 
                    end
            in (buildlet, livein colis) end             
          | comp (CaseS(e0, alts))   =
            (* We assume all alternatives are present and in the right order *)
            let val (co0,  li0)  = comp e0
                val (coas, lias) = compilealts alts
            in (fn ce => Case(co0 ce, coas ce), li0 ++ lias) end
          | comp (CstS (Int i))  = (fn ce => Cst i, emptyset)
          | comp (CstS (Bool b)) = 
		let val i = if b then 1 else 0
		in (fn ce => Cst i, emptyset) seq (prim Mkbool) end
          | comp (Op2S (III opr, e1, e2)) = 
            (comp e1) seq (comp e2) seq (prim (Op2 opr))
          | comp (Op2S (IIB opr, e1, e2)) = 
            (comp e1) seq (comp e2) seq (prim (Op2 opr)) seq (prim Mkbool)
          | comp (IfS(e0, e1, e2))    =
            let val (co0,  li0)  = comp e0
                val (coas, lias) = compilealts [(1, [], e2), (2, [], e1)]
            in (fn ce => Case(co0 ce, coas ce), li0 ++ lias) end
        val (co, _) = comp e 
    in co [] end

end
