(* Run -- the abstract machine itself. *)

structure Run :> Run = 
struct
    open Data;

    type 'a pointer = ('a option * bool) ref

    datatype 'e closure = Clos of 'e * environment
    and environment = 
        Env of expr closure pointer list
      | Bas of int

    datatype stkobj = 
        Mrk of expr closure pointer             (* update marker *)
      | Arg of expr closure pointer             (* argument      *)
      | Cse of (expr * trim) list closure       (* case marker   *);

    val memsizelam = ref false;          (* print heap size at application *)
    val memsizeprt = ref false;          (* print heap size after printing *)

    local
	(* Count the reachable heap cells.  Invariant: Between calls
        to memsize, every reachable heap cell is either (NONE, _) or
        (_, !visited) *)

        val visited = ref false;
        fun reach (r as ref (NONE, _), count)  = count
          | reach (r as ref (vopt as (SOME (Clos (_, E))), state), count) =
            if state = !visited then count
            else (r := (vopt, !visited); reachenv(E, count+1))
        and reachenv (Bas _, count) = count
          | reachenv (Env E, count) = foldr reach count E
        fun reachstkobj (Mrk _, count)             = count
          | reachstkobj (Arg r, count)             = reach(r, count)
          | reachstkobj (Cse (Clos (_, E)), count) = reachenv(E, count)
        fun reachstk (S, count) = foldr reachstkobj count S
    in
        exception Blackhole
        fun allocate ()   = ref (NONE, !visited)
        fun delete r      = r := (NONE, !visited)
        fun update (r, v) = r := (SOME v, !visited)
        fun access r      = case !r of
                               (NONE,   _) => raise Blackhole
                             | (SOME v, _) => v
        fun memsize (E, S) = 
            let val _ = visited := not (!visited)
                val heap = reachenv(E, reachstk(S, 0)) 
            in print ("{" ^ Int.toString (heap + length S) ^ "}") end
    end;

    fun trim (E, trim) =
        let fun h (j, _, []) = []
              | h (j, [], _) = raise Impossible "Run.trim.h"
              | h (j, i::ir, ts as t1 :: tr) =
                if j=t1 then i :: h(j+1, ir, tr)
                        else      h(j+1, ir, ts)
        in h(1, E, trim) end

    fun exec (App (e, i), Env E, V, S)  = 
        exec(e, Env E, V, Arg (List.nth (E, i-1)) :: S)
      | exec (Lam e, Env E, V, Arg p :: S) = 
        (if !memsizelam then memsize(Env E, Arg p::S) else ();
         exec(e, Env (p :: E), V, S))
      | exec (Var i, Env E, V, S)       =   
        let val p = List.nth (E, i-1)
            val Clos (ep, Ep) = access p
        in delete p; exec(ep, Ep, V, Mrk p :: S) end
      | exec (Lam e, E, V, Mrk p :: S)  =   
        (update(p, Clos(Lam e, E)); exec (Lam e, E, V, S))
      | exec (Let(ets, (e0, t0)), Env E, V, S) =
        let val pets = map (fn (e, t) => (allocate (), e, t)) ets
            val E'   = map #1 pets @ E
        in  
            ignore 
	    (map (fn (p,e,t) => update(p, Clos(e, Env(trim(E',t))))) pets);
            exec (e0, Env(trim(E',t0)), V, S) 
        end
      | exec (Case(e, (alts, t)), Env E, V, S)   = 
        exec(e, Env E, V, Cse (Clos (alts, Env (trim (E, t)))) :: S)
      | exec (Con(k,a), Env E', V, Cse (Clos (alts, Env E)) :: S) =
        let val (ek, tk) = List.nth(alts, k-1)
        in exec(ek, Env (trim (E' @ E, tk)), V, S) end
      | exec (Con(k,a), E, V, Mrk p::S) =
        (update(p, Clos(Con(k,a), E)); 
         exec (Con(k,a), E, V, S))
      | exec(Ret, bas as Bas b, V, Mrk p :: S) =
        (update(p, Clos(Ret, bas)); exec (Ret, bas, V, S))
      | exec (Ret, Bas b, V, Cse(Clos([(e,_)],E)) :: S) =
        exec(e, E, b :: V, S)
      | exec (Cst b, _, V, S)           = exec(Ret, Bas b, V, S)
      | exec (Op2 opr, _, b2::b1::V, S) =
        exec(Ret, Bas(opr(b1,b2)), V, S)
      | exec (Mkbool, E, b :: V, S)   =
        exec(if b=0 then Con(1,0) else Con(2,0), E, V, S)
      | exec (Prtint, _, i :: V, Cse(Clos([(e, _)],E)) :: S) =
        (print (Int.toString i); 
         if !memsizeprt then memsize(E, S) else ();
         exec(e, E, V, S))
      | exec (Prtstr s, _, V, Cse(Clos([(e,_)],E)) :: S) =
        (print s; exec(e, E, V, S))
      | exec (e, E, V, S)               = (e, E, V, S)

    type environment = environment
    and stkobj = stkobj;

    fun run e = let val res = exec(e, Env [], [], []) in print "\n"; res end
    val memsizeprt = memsizeprt
    val memsizelam = memsizelam
end
